#!/usr/bin/python -u
#
# Client code for Update Agent
# Copyright (c) 1999-2001 Red Hat, Inc.  Distributed under GPL.
#
# Author: Preston Brown <pbrown@redhat.com>

import os,sys
import string
import re
import fnmatch, glob
import rpm
import socket
import cgiwrap
import xmlrpclib
import time
import StringIO

sys.path.append("/usr/share/rhn/up2date")

from translate import _, N_, cat

import config

cfg = config.Up2dateConfig()
cfg.load()

def version():
    # substituted to the real version by the Makefile at installation time.
    return "2.5.4"


DEBUG=1
def DPRINT(string):
    if DEBUG:
        print  string

# get around cached dependencies.  Gross ugly hack.
rpm.addMacro("_dbi_tags","Packages:Name:Basenames:Group:Requirename:Providename:Conflictname:Triggername")

# this is a "gloabal" variable caching the list of all available packages
global avail_package_lst
avail_package_lst = []


# attempt to log all interesting stuff, namely, anything that hits the network
# any error messages, package installs, etc
class Log:
    def __init__(self):
        self.app = "up2date"
        if cfg.readEntry("logFile"):
            self.log_file_name = cfg.readEntry("logFile")
        else:
            self.log_file_name = "/var/log/up2date"

        self.log_info = "[%s]" % time.ctime(time.time())
        
    def log_me(self, *args):
        if len(args) > 0:
            s = "%s " % self.app
            for i in args:
                s = s + "%s" % i
        else:
            s = "%s " % (self.app)
        self.write_log(s)

    def trace_me(self):
        import traceback
        x = traceback.extract_stack()
        bar = string.join(traceback.format_list(x))
        self.write_log(bar)

    def write_log(self, s):
        self.log_file = open(self.log_file_name, 'a')
        msg = "%s %s\n" % (self.log_info, str(s))
        self.log_file.write(msg)
        self.log_file.flush()
        self.log_file.close()
        # very simple "verbose" mode. Need a real verbose mode
        # turned off by preston 02/22/01
        #if cfg.readEntry("verbose"):
        #    sys.stdout.write("%s\n" % str(s))
        #   sys.stdout.flush()
        
log = Log()


class Error:
    # base class for client errors
    def __init__(self, errmsg):
        self.errmsg = errmsg

    def __repr__(self):
        log.log_me(self.errmsg)
        return self.errmsg
    
class FileError(Error):
    def __repr__(self):
        msg = _("Disk error.  The message was:\n") + self.errmsg
        log.log_me(msg)
        return msg

class RpmError(Error):
    def __repr__(self):
        msg = _("RPM error.  The message was:\n") + self.errmsg
        log.log_me(msg)
        return msg

class ConflictError(Error):
    def __repr__(self):
        msg = _("RPM conflict error.  The message was:\n") + self.errmsg
        log.log_me(msg)
        return msg

class DependencyError(Error):
    def __repr__(self):
        msg = _("RPM dependency error.  The message was:\n") + self.errmsg
        log.log_me(msg)
        return msg

class CommunicationError(Error):
    # indicates a problem doing xmlrpc communication
    def __repr__(self):
        msg =  _("Error communicating with server.  The message was:\n") + self.errmsg
        log.log_me(msg)
        return msg

class DelayError(Error):
    # indicates a problem doing xmlrpc communication
    def __repr__(self):
        msg =  _("Delay error from server.  The message was:\n") + self.errmsg
        log.log_me(msg)
        return msg

def getArch():
    s = os.uname()[4]
    return s

# return a glob for your particular architecture.
def archGlob():
    if re.search("i.86", os.uname()[4]):
        return "i?86"
    elif re.search("sparc", os.uname()[4]):
        return "sparc*"
    else:
        return os.uname()[4]

# small class for header caching pretending to be a dict
# in theory we could eventually make this smart and stuff
# (ie, mru, etc)
#
# implement it this way for now so we can make a max header
# cache size
#
# Need a miniHeaderCache, which caches all headers
# but only: name,version,release,arch,summary,description,size
# Otherwise the size computations are slow as dirt
#
class HeaderCache:
    def __init__(self):
        # how many headers to cache in ram
        if cfg.readEntry("headerCacheSize"):
            self.cache_size = cfg.readEntry("headerCacheSize")
        else:
            self.cache_size = 30
        self.__cache = {}
        self.__cacheLite = {}
        
    def set_cache_size(self, number_of_headers):
        self.cache_size = number_of_headers

    def __liteCopy(self, header):
        tmp = {}
        tmp['name'] = header['name']
        tmp['version'] = header['version']
        tmp['release'] = header['release']
        tmp['arch'] = header['arch']
        tmp['summary'] = header['summary']
        tmp['description'] = header['description']
        tmp['size'] = header['size']
        return tmp
        
    def __setitem__(self,item,value):
        if len(self.__cache) <= self.cache_size:
            self.__cache[item] = value
            self.__cacheLite[item] = self.__liteCopy(value)
        else:
            # okay, this is about as stupid of a cache as you can get
            # but if we hit the max cache size, this is as good as
            # any mechanism for freeing up space in the cache. This
            # would be a good place to put some smarts
            bar = self.__cache.keys()
            del self.__cache[bar[self.cache_size-1]]
            self.__cache[item] = value
            self.__cacheLite[item] = self.__liteCopy(value)

    def __getitem__(self, item):
        return self.__cache[item]

    def getLite(self, item):
        return self.__cacheLite[item]
    
    def __len__(self):
        return len(self.__cache)

    def keys(self):
        return self.__cache.keys()

    def values(self):
        return self.__cache.keys()
    
    def has_key(self, item,lite=None):
        if lite:
            return self.__cacheLite.has_key(item)
        else:
            return self.__cache.has_key(item)

    def __delitem__(self, item):
        del self.__cache[item]

    def printLite(self):
        print self.__cacheLite

headerCache = HeaderCache()

def getVersion():
    ret = ""

    if cfg.readEntry("versionOverride") != "":
        return cfg.readEntry("versionOverride")
        
    db = openrpmdb()
    try:
        index = db.findbyprovides("redhat-release")[0]
    except:
        raise RpmError("Couldn't determine what version of Red Hat Linux you are running.\nIf you get this error, try running \n\n\t\trpm --rebuilddb\n\n")
    return (db[index])['version']


# mainly here to make conflicts resolution cleaner
def findDepLocal(db, dep):
    header = None
    if dep[0] == '/':
        # Treat it as a file dependency
        try:
            hdr_arry = db.findbyfile(dep)
        except:
            hdr_arry = []
            
        for n in hdr_arry:
            header = db[n]
            break
    else:
        # Try it first as a package name
        try:
            hdr_arry = db.findbyname(dep)
        except:
            hdr_arry = []
        for n in hdr_arry:
            header = db[n]
            break
        else:
            # else try it as a soname
            try:
                hdr_arry = db.findbyprovides(dep)
            except:
                hdr_arry = []
            for n in hdr_arry:
                header = db[n]
                break
            
    if header != None:
        return 1
    else:
        return 0
    

def getSystemId():
    path = cfg.readEntry("systemIdPath")
    if not os.access(path, os.R_OK):
        return None
    
    f = open(path, "r")
    ret = f.read()
        
    f.close()
    return ret

def hasSSL():
    return hasattr(socket, "ssl")

def getServer(refreshCallback=None):

    if cfg.readEntry("enableProxy"):
        proxyHost = cfg.readEntry("httpProxy")
    # legacy name for setting, read this as a 2nd attempt
        if proxyHost == "":
            try:
                proxyHost = cfg.readEntry("pkgProxy")
            except:
                proxyHost = None

    else:
        proxyHost = None
        
    if hasSSL():
        serverUrl = cfg.readEntry("serverURL")
    else:
        serverUrl = cfg.readEntry("noSSLServerURL")

    if cfg.readEntry("enableProxyAuth"):
        if cfg.readEntry("proxyUser"):
            proxyUser = cfg.readEntry("proxyUser")
            
        if cfg.readEntry("proxyPassword"):
            proxyPassword = cfg.readEntry("proxyPassword")

    else:
        proxyUser = None
        proxyPassword = None

    lang = None
    for env in 'LANGUAGE', 'LC_ALL', 'LC_MESSAGES', 'LANG':
        if os.environ.has_key(env):
            if not os.environ[env]:
                # sometimes unset
                continue
            lang = string.split(os.environ[env], ':')[0]
            lang = string.split(lang, '.')[0]
            break

        
    s = xmlrpclib.Server(serverUrl, proxyHost,
                         refreshCallback=refreshCallback,
                         username=proxyUser,
                         password=proxyPassword)
    if lang:
        s.setlang(lang)
    
    return s


# if a user has upgraded to a newer release of Red Hat but still
# has a systemid from their older release, they need to get an updated
# systemid from the RHN servers.  This takes care of that.
def maybeUpdateVersion():
    try:
        idVer = xmlrpclib.loads(getSystemId())[0][0]['os_release']
    except:
        # they may not even have a system id yet.
        return 0

    systemVer = getVersion()

    if idVer != systemVer:
      s = getServer()
    
      try:
          newSystemId = s.registration.upgrade_version(getSystemId(),
                                                       systemVer)
      except KeyboardInterrupt:
          raise CommunicationError(_("Connection aborted by the user"))
      except (socket.error, socket.sslerror), e:
          if len(e.args) > 1:
              raise CommunicationError(e.args[1])
          else:
              raise CommunicationError(e.args[0])
      except cgiwrap.ProtocolError, e:
          raise CommunicationError(e.errmsg)
      except xmlrpclib.ResponseError:
        raise CommunicationError("Broken response from the server.");
      except xmlrpclib.Fault, f:
          raise CommunicationError(f.faultString)

      path = cfg.readEntry("systemIdPath")
      dir = path[:string.rfind(path, "/")]
      if not os.access(dir, os.W_OK):
          try:
              os.mkdir(dir)
          except:
              return 0
      if not os.access(dir, os.W_OK):
          return 0

      if os.access(path, os.F_OK):
          # already have systemid file there; let's back it up
          savePath = path + ".save"
          try:
              os.rename(path, savePath)
          except:
              return 0

      f = open(path, "w")
      f.write(newSystemId)
      f.close()
      try:
          os.chmod(path, 0600)
      except:
          pass
      

def checkGPGInstallation():
    if not os.access("/usr/bin/gpg", os.X_OK):
        log.log_me("error: GPG is not installed")
        # not installed or accessible.
        return 1

    output = os.popen("/usr/bin/gpg --list-keys 2>/dev/null", "r").read()
    if not re.search("pub  1024D/DB42A60E 1999-09-23 Red Hat, Inc <security@redhat.com>", output):
        return 2

    return 0

def addGPGKey():
    # might as well log this security stuff well
    log.log_me("Adding /usr/share/rhn/RPM-GPG-KEY to keyring")
    log.log_me("/usr/bin/gpg --import /usr/share/rhn/RPM-GPG-KEY >/dev/null 2>&1")
    return os.system("/usr/bin/gpg --import /usr/share/rhn/RPM-GPG-KEY >/dev/null 2>&1")

def retrieveOnly():
    return cfg.readEntry("retrieveOnly")

def openrpmdb(option=0, dbpath=None):
    if not dbpath:
        if cfg.readEntry("dbpath"):
            dbpath = cfg.readEntry("dbpath")
        else:
            dbpath = "/var/lib/rpm/"

    rpm.addMacro("_dbpath", dbpath)

    log.log_me("Opening rpmdb in %s with option %s" % (dbpath,option))
    try:
        db = rpm.opendb(option)
    except rpm.error, e:
        raise RpmError(_("Could not RPM database for reading.  Perhaps it is already in use?"))

    return db
        
def installedHeader(someName,db):
    h = None

    if type(someName) == type([]):
        pkgName = someName[0]
    else:
        pkgName = someName

# FIXME UGLY KLUGE FOR RPM BINDING weirdness in 7.0
    try:
	bar = db.findbyname(pkgName)
    except: 
	bar = []

    try:
        for index in bar:
            h = db[index]
            break
    except rpm.error, e:
        raise RpmError(e)

    if h != None:
        name = h['name']
        epoch = h['epoch']
        if epoch == None:
            epoch = ""
        version = h['version']
        release = h['release']
        if type(someName) == type([]):
            if (pkgName == name and
                pkgName[2] == version and
                pkgName[3] == release and
                pkgName[4] == epoch):
                return h
        else:
            if (pkgName == name):
                return h

    # not found
    return None

def pkgToString(pkg):
    return "%s-%s-%s" % (pkg[0], pkg[1], pkg[2])

# compare two RPM packages
def comparePackages(pkgLabel1, pkgLabel2):
    name1 = pkgLabel1[0]
    version1 = pkgLabel1[1]
    release1 = pkgLabel1[2]
    epoch1 = pkgLabel1[3]
    name2 = pkgLabel2[0]
    version2 = pkgLabel2[1]
    release2 = pkgLabel2[2]
    epoch2 = pkgLabel2[3]

    if epoch1 == "":
        epoch1 = None
    else:
        epoch1 = "%s" % epoch1
        
    if epoch2 == "":
            epoch2 = None
    else:
        epoch2 = "%s" % epoch2
        
    return rpm.labelCompare((epoch1, version1, release1),
                            (epoch2, version2, release2))


def md5(fileName):
    import md5
    hash = md5.new()
    
    try:
        f = open(fileName, "r")
    except:
        return ""

    fData = f.read()
    hash.update(fData)
    del fData
    f.close()
    
    hex = string.hexdigits
    md5 = ""
    for c in hash.digest():
        i = ord(c)
        md5 = md5 + hex[(i >> 4) & 0xF] + hex[i & 0xf]

    return md5
    
# has modified config files.  Returns a list of filenames
# Not used as of 03/01/2001. Remove?
def modifiedConfigs(header):
    fileNames = header['fileNames']
    fileFlags = header['fileflags']
    md5Sums = header['filemd5s']
    modified = []
    
    for f_i in range(len(fileNames)):
        if fileFlags[f_i] & rpm.RPMFILE_CONFIG:
            if md5Sums[f_i] != md5(fileNames[f_i]):
                modified.append(fileNames[f_i])

    return modified

# returns a list of config filenames from a header
# Not used as of 03/01/2001.  Remove?
def configFiles(header):
    fileNames = header['filenames']
    fileFlags = header['fileflags']
    configs = []
    
    for f_i in range(len(fileNames)):
        if fileFlags[f_i] & rpm.RPMFILE_CONFIG:
            configs.append(fileNames[f_i])

    return configs


def getAvailablePackageList():
    global avail_package_lst

    # if this is non-zero, we've already fetched it
    if len(avail_package_lst):
        return avail_package_lst

    # avail_package_list)
    availPkgList = []

    # need a PackageSourceList to list all the places we
    # can potentially get packages from. Then we would iterate
    # though each of them, and build a composite list. 
    # this would be a list of packageSource objects
    # -akl-
    s = getServer()

    try:
        log.log_me("getAvailablePackageList from network")
    except:
        pass
    
    try:
        availPkgList = s.up2date.listall_size(getSystemId())
    except KeyboardInterrupt:
        raise CommunicationError(_("Connection aborted by the user"))
    except (socket.error, socket.sslerror), e:
        if len(e.args) > 1:
            raise CommunicationError(e.args[1])
        else:
            raise CommunicationError(e.args[0])
    except cgiwrap.ProtocolError, e:
        raise CommunicationError(e.errmsg)
    except xmlrpclib.ResponseError:
        raise CommunicationError("Broken response from the server.");
    except xmlrpclib.Fault, f:
        raise CommunicationError(f.faultString)

    avail_package_lst = availPkgList
    return availPkgList
    

def getPackageSearchPath():
    dir_list = []
    dir_list.append(cfg.readEntry("storageDir"))

    dir_string = cfg.readEntry("packageDir")
    if dir_string: 
    	dir_list = dir_list + string.split(dir_string, ':')
    return dir_list

# pass the lite flag if you only want the "lite" header
def getHeader(pkg, lite=None):
    hdr = None

    # first see if it in memory cache:
    if lite:
        if headerCache.has_key(pkgToString(pkg),lite=1):
            return headerCache.getLite(pkgToString(pkg))

    if headerCache.has_key(pkgToString(pkg)):
        return headerCache[pkgToString(pkg)]
    
    hdr = getHeaderOffDisk(pkg)
    if hdr:
        return hdr

    s = getServer()
    
    try:
        ret = s.up2date.header(getSystemId(), pkg)
    except KeyboardInterrupt:
        raise CommunicationError(_("Connection aborted by the user"))
    except (socket.error, socket.sslerror), e:
        if len(e.args) > 1:
            raise CommunicationError(e.args[1])
        else:
            raise CommunicationError(e.args[0])
    except cgiwrap.ProtocolError, e:
        raise CommunicationError(e.errmsg)
    except xmlrpclib.ResponseError:
        raise CommunicationError("Broken response from the server.");
    except xmlrpclib.Fault, f:
        raise CommunicationError(f.faultString)

    bin = ret[0]
    hdr = rpm.headerLoad(bin.data)
    saveHeader(hdr)
    headerCache["%s-%s-%s" % (hdr['name'],
                              hdr['version'],
                              hdr['release'])] = hdr
    return hdr

def getHeaderOffDisk(pkg, msgCallback = None, progressCallback = None):
    hdr = None

    dir_list = getPackageSearchPath()
    for dir in dir_list:
        fileNames = glob.glob("%s/%s.%s.*" %
                              (dir,
                               pkgToString(pkg),
                               archGlob()))
        # if we find anything, bail
        if len(fileNames):
            break

    # since we found a match, now see if its a header or an rpm
    if len(fileNames):
        if os.access(fileNames[0], os.R_OK):
            if not re.search("rpm$", fileNames[0]):
                # it wasnt an rpm, so must be a header
                if os.stat(fileNames[0])[6] == 0:
                    return None
                fd = open(fileNames[0], "r")
                # if this header is corrupt, rpmlib exits and we stop ;-<
                try:
                    hdr = rpm.headerLoad(fd.read())
                except:
                    return None
                headerCache[pkgToString(pkg)] = hdr
                fd.close()
                return hdr
            else:
                if not checkRpmMd5(fileNames[0]):
                    fd = os.open(fileNames[0], 0)
                    try:
                        (hdr,isSource) = rpm.headerFromPackage(fd)
                    except:
                        os.close(fd)
                        raise RpmError(_("Error reading header"))
                    os.close(fd)
                    log.log_me("Read header from %s" % fileNames[0])
                    headerCache[pkgToString(pkg)] = hdr
                    return hdr
                else:
                    return None

        else:
            return None
    else:
        for dir in dir_list:
            fileNames = glob.glob("%s/%s.noarch.*" %
                                      (dir,
                                       pkgToString(pkg)))
                        
        if len(fileNames):
            if os.access(fileNames[0], os.R_OK):
                if not re.search("rpm$", fileNames[0]):
                    # it's not an rpm, must be a header
                    if os.stat(fileNames[0])[6] == 0:
                        return None
                    fd = open(fileNames[0], "r")
                    try:
                        hdr = rpm.headerLoad(fd.read())
                    except:
                        log.log_me("Corrupt header %s, skipping, will download later..." % fileNames[0])
                        return None
                    fd.close()
                    headerCache[pkgToString(pkg)] = hdr
                    return hdr
                else:
                    if not checkRpmMd5(fileNames[0]):
                        fd = os.open(fileNames[0],0)
                        try:
                            (hdr,isSource) = rpm.headerFromPackage(fd)
                        except:
                            os.close(fd)
                            raise RpmError(_("Error reading header"))
                        os.close(fd)
                        log.log_me("Reading header from: %s" % fileNames)
                        headerCache[pkgToString(pkg)] = hdr
                        return hdr
                    else:
                        return None
            else:
                return None
        else:
            return None


# implicitly places headers in the headerCache.
def getHeaders(pkgList, msgCallback = None, progressCallback = None,
               refreshCallback = None):

    notOnDiskList = []
    headerList = []
    # -akl-
    # package source object would be used here too, need a "header"
    # method

        
    # if we get a single header, listify it
    # double check to make sure it's a list of lists and not just
    # a single list item
    if type(pkgList) == type([]):
        if len(pkgList):
            if type(pkgList[0]) == type(""):
                tmp = [pkgList]
                pkgList = tmp

    total = len(pkgList)

    # reset progress bar
    if progressCallback:
        progressCallback(0, total)

    if not refreshCallback:
        refreshCallback = 0

    for pkg in pkgList:
        if refreshCallback:
            refreshCallback()
        hdr = None
        foundHeader = None
        # check if its already in the memory cache
        if headerCache.has_key(pkgToString(pkg)):
            hdr = headerCache[pkgToString(pkg)]
            headerList.append(hdr)
            foundHeader = 1
            if progressCallback:
                progressCallback(len(headerList), total)
            continue
            # check to see if its on disk
        hdr = getHeaderOffDisk(pkg)
        if hdr:
            headerCache[pkgToString(pkg)] = hdr
            headerList.append(hdr)
            foundHeader=1
            if progressCallback:
                progressCallback(len(headerList), total)

        if not foundHeader:
            notOnDiskList.append(pkg)
        

    # any remaining ones we need to get from the server
    # this fetches a pile of headers at once. If we were evil
    # we could extend this and spin off a few threads and get
    # multiple headers in parallel. But lets be nice to the servers
    # for the time being
    if len(notOnDiskList):
        if cfg.readEntry("headerFetchCount"):
            headerFetchCount = cfg.readEntry("headerFetchCount")
        else:
            # a reasonable default
            headerFetchCount = 10
        last_inc = 0
        index = 0
        # lame hack cause we cant marshall "None"
        if not refreshCallback:
            refreshCallback = 0
        while index <= range(total):
            s = getServer(refreshCallback=refreshCallback)
            if (total - index < headerFetchCount):
                next_inc = last_inc + (total%headerFetchCount)
            else:
                next_inc = last_inc + headerFetchCount

            listToFetch = notOnDiskList[last_inc:next_inc]
            if len(listToFetch) == 0:
                break

            # turn on compression
            #s.set_transport(2, 2)

            # log stuff we actually hit the network for
            try:
                log.log_me("GetHeaders from network: %s" % map(lambda a : "%s-%s-%s" % (a[0],a[1],a[2]), listToFetch))
            except:
                pass

            try:
                ret = s.up2date.header(getSystemId(), listToFetch)
            except KeyboardInterrupt:
                raise CommunicationError(_("Connection aborted by the user"))
            except (socket.error, socket.sslerror), e:
                if len(e.args) > 1:
                    raise CommunicationError(e.args[1])
                else:
                    raise CommunicationError(e.args[0])
            except cgiwrap.ProtocolError, e:
                raise CommunicationError(e.errmsg)
            except xmlrpclib.ResponseError:
                raise CommunicationError("Broken response from the server.");
            except xmlrpclib.Fault, f:
                raise CommunicationError(f.faultString)

            for bin in ret:
                hdr = rpm.headerLoad(bin.data)
                saveHeader(hdr)
                headerCache["%s-%s-%s" % (hdr['name'],
                                      hdr['version'],
                                      hdr['release'])] = hdr
                headerList.append(hdr)


            if progressCallback:
                progressCallback(len(headerList), total)
            
            last_inc = next_inc
            index = next_inc
            if index > total:
                break

    return headerList


def solveDep(unknowns,refreshCallback=None):
    retList = []

    if not refreshCallback:
        refreshCallback = 0

    s = getServer(refreshCallback=refreshCallback)

    try:
        retList = s.up2date.solvedep(getSystemId(), unknowns)
    except KeyboardInterrupt:
        raise CommunicationError(_("Connection aborted by the user"))
    except (socket.error, socket.sslerror), e:
        if len(e.args) > 1:
            raise CommunicationError(e.args[1])
        else:
            raise CommunicationError(e.args[0])
    except cgiwrap.ProtocolError, e:
        raise CommunicationError(e.errmsg)
    except xmlrpclib.ResponseError:
        raise CommunicationError("Broken response from the server.");
    except xmlrpclib.Fault, f:
        if f.faultCode == -26:
            #raise RpmError(f.faultString + _(", depended on by %s") % unknowns)
            raise RpmError(f.faultString)
        else:
            raise CommunicationError(f.faultString)

    if len(retList):
	availList = getAvailablePackageList()
	availList.sort()
	newList = []
        availListNames = map(lambda p: p[0], availList)       
	for pkg in retList:
            if pkg[0] in availListNames:                
                newList.append(availList[availListNames.index(pkg[0])])
            else:
                newList.append(pkg)
	retList = newList

    return retList


def saveHeader(hdr):
    fileName = "%s/%s.%s.hdr" % (cfg.readEntry("storageDir"),
                              string.join((hdr['name'],
                                           hdr['version'],
                                           hdr['release']), "-"), hdr['arch'])

    fd = os.open(fileName, os.O_WRONLY|os.O_CREAT, 0600)

    os.write(fd, hdr.unload())
    os.close(fd)

    return 1

    
def removeHeader(pkg):
    # first see if it is on disk:
    fileNames = glob.glob("%s/%s.%s.hdr" %
                          (cfg.readEntry("storageDir"),
                           pkgToString(pkg),
                           archGlob()))
    for fileName in fileNames:
        try:
            os.remove(fileName)
        except:
            raise FileError(_("cannot remove %s.") % fileName)

    fileNames = glob.glob("%s/%s.noarch.hdr" %
                          (cfg.readEntry("storageDir"),
                           pkgToString(pkg)))
    for fileName in fileNames:
        try:
            os.remove(fileName)
        except:
            raise FileError(_("cannot remove %s.") % fileName)

def checkRpmMd5(fileName):
    # FIXME
    #insert magic to "shutup" rpm here 
    # pgb 03/01/2001: not sure if any magic can fix that in RPM. :(
    
    # only newer RPMs allow this
    if "checksig" in dir(rpm):
        # RPM spews to stdout/stderr.  Redirect.
        saveStdout = os.dup(1)
        saveStderr = os.dup(2)
        redirStdout = os.open("/dev/null", os.O_WRONLY | os.O_APPEND)
        redirStderr = os.open("/dev/null", os.O_WRONLY | os.O_APPEND)
        os.dup2(redirStdout, 1)
        os.dup2(redirStderr, 2)
        # now do the rpm thing
        ret = rpm.checksig(fileName, rpm.RPMTAG_SIGMD5)
        # restore normal stdout and stderr
        os.dup2(saveStdout, 1)
        os.dup2(saveStderr, 2)
        # Clean up
        os.close(redirStdout)
        os.close(redirStderr)
        os.close(saveStdout)
        os.close(saveStderr)
        return ret
    else:
        # legacy 3.0.x RPM
        sigOk = 0
        lines = os.popen("/bin/rpm -Kv --nopgp --nogpg %s 2>&1" % fileName, "r").read()
        if not re.search("MD5 sum OK:", lines):
            sigOk = 1

        return sigOk


def getPackage(pkg, msgCallback = None, progressCallback = None):
    hdr = getHeader(pkg)
    count = 0
    baseFileName = "%s.%s.rpm" % (string.join((hdr['name'],
                                           hdr['version'],
                                           hdr['release']), "-"),
                              hdr['arch'])

    if msgCallback != None:
        msgCallback("%s" % baseFileName, hdr['name'])

    fileName = "%s/%s.%s.rpm" % (cfg.readEntry("storageDir"),
                                 string.join((hdr['name'],
                                              hdr['version'],
                                              hdr['release']), "-"),
                                 hdr['arch'])
    # look for the file in all the packageDir paths, and symlink it to
    # storeageDir to make life easy
    #
    # FIXME: make sure the symlink cleanup handling is okay
    # FIXME: Is this a security risk? It is a package the admin told us to
    # install, and it does have to go through the same sanity checks.
    dir_list = getPackageSearchPath()
    for dir in dir_list:
        tmpFileNames = glob.glob("%s/%s" % (dir, baseFileName))
        # if the file is in the storageDir, dont symlink it to itself ;->
        if len(tmpFileNames) and not (dir == cfg.readEntry("storageDir")):
            try:
                os.remove(fileName)
            except OSError:
                pass
            os.symlink(tmpFileNames[0], fileName)
            break
    
    # check if we already have package and that they are valid
    if os.access(fileName, os.R_OK) and not checkRpmMd5(fileName):
        if progressCallback != None:
            progressCallback(1,1)

	# akl-  create a packageSource object? this would include 
	# methods and data about where the package is, and how to
	# download it. The chunking method for rhn packages that
	# follows would be an example
	# fetching a file off a fs would be another
	# use the fifth field in the availPackageList to determine
	# which object to use
    else:
        log.log_me("Downloading package: %s-%s-%s" % (pkg[0],pkg[1],pkg[2]) )
        s = getServer()
        try:
            pkgStream = s.up2date.package(getSystemId(), pkg)
        except KeyboardInterrupt:
            raise CommunicationError(_("Connection aborted by the user"))
        except (socket.error, socket.sslerror), e:
            if len(e.args) > 1:
                raise CommunicationError(e.args[1])
            else:
                raise CommunicationError(e.args[0])
        except cgiwrap.ProtocolError, e:
            raise CommunicationError(e.errmsg)
        except xmlrpclib.ResponseError:
            raise CommunicationError("Broken response from the server.");
        except xmlrpclib.Fault, f:
            raise CommunicationError(f.faultString)

        try:
            fd = os.open(fileName, os.O_WRONLY|os.O_CREAT|os.O_TRUNC, 0600)
        except:
            print _("Error opening output file %s for writing.") % fileName
            return 0

        startTime = time.time()
        lastTime = startTime
        chunk = pkgStream.read(1024)
        while len(chunk) > 0:
            curTime = time.time()
            try:
                os.write(fd, chunk)
            except:
                os.close(fd)
                print _("Error writing to %s") % fileName
                return 0

            # handle progress, speed, and time.
            count = count + len(chunk)

            if curTime - lastTime >= 1 or count == len(pkgStream):
                speed = count / (curTime - startTime)
                lastTime = curTime
                if speed != 0:
                    if len(pkgStream) == 0:
                        secs = 0
                    else:
                        secs = (len(pkgStream) - count) / speed
                if progressCallback != None:
                    progressCallback(count, len(pkgStream),
                                     speed, secs)
                    
            chunk = pkgStream.read(1024)

        os.close(fd)

    if cfg.readEntry("retrieveSource"):
        s = getServer()
        try:
            pkgStream = s.up2date.source_package_by_name(getSystemId(),
                                                         hdr['SOURCERPM'])
        except KeyboardInterrupt:
            raise CommunicationError(_("Connection aborted by the user"))
        except (socket.error, socket.sslerror), e:
            if len(e.args) > 1:
                raise CommunicationError(e.args[1])
            else:
                raise CommunicationError(e.args[0])
        except cgiwrap.ProtocolError, e:
            raise CommunicationError(e.errmsg)
        except xmlrpclib.ResponseError:
            raise CommunicationError("Broken response from the server.");
        except xmlrpclib.Fault, f:
            raise CommunicationError(f.faultString)

        fileName = pkgStream.name
        if fileName == "":
            # if the database isn't properly populated we get an empty string
            return 1

        if msgCallback != None:
            msgCallback("%s" % fileName, hdr['name'])

        fileName = "%s/%s" % (cfg.readEntry("storageDir"), fileName)

        # check if we already have package
        if os.access(fileName, os.R_OK):
            if progressCallback != None:
                progressCallback(1, 1)

            return 1

        try:
            # at some point we probabaly want to do a "reget" style support here
            fd = os.open(fileName,os.O_WRONLY|os.O_CREAT|os.O_TRUNC, 0600)
        except:
            print _("Error opening output file %s for writing.") % fileName
            return 0

        startTime = time.time()
        lastTime = startTime
        chunk = pkgStream.read(1024)
        while len(chunk) > 0:
            curTime = time.time()
            try:
                os.write(fd, chunk)
            except:
                os.close(fd)
                print _("Error writing to %s") % fileName
                return 0

            count = count + len(chunk)

            if curTime - lastTime >= 1 or count == len(pkgStream):
                speed = count / (curTime - startTime)
                lastTime = curTime
                if speed != 0:
                    if len(pkgStream) == 0:
                        secs = 0
                    else:
                        secs = (len(pkgStream) - count) / speed
                if progressCallback != None:
                    progressCallback(count, len(pkgStream),
                                     speed, secs)

            chunk = pkgStream.read(1024)

        os.close(fd)

    return 1


def removePackage(pkg):
     # first see if it is on disk:
    fileNames = glob.glob("%s/%s.%s.*" %
                          (cfg.readEntry("storageDir"),
                           pkgToString(pkg),
                           archGlob()))
    for fileName in fileNames:
        log.log_me("deleting %s" % fileName)
        try:
            os.remove(fileName)
            
        except:
            raise FileError(_("Cannot remove %s") % fileName)
        
    fileNames = glob.glob("%s/%s.noarch.*" %
                          (cfg.readEntry("storageDir"),
                           pkgToString(pkg)))
    for fileName in fileNames:
        log.log_me("deleting %s" % fileName)
        try:
            os.remove(fileName)
        except:
            raise FileError(_("Cannot remove %s") % fileName)


# Not used as of 03/01/2001.  Remove?
def getPackageInfo(pkg):
    s = getServer()
    log.log_me("getPackageInfo for %s" % pkg)
    try:
        info = s.up2date.description(getSystemId(), [pkg])
    except KeyboardInterrupt:
        raise CommunicationError(_("Connection aborted by the user"))
    except xmlrpclib.ResponseError:
        raise CommunicationError("Broken response from the server.");
    except xmlrpclib.Fault, f:
        print f
        return None

    return info


def getAdvisoryInfo(pkg):
    s = getServer()

    db = openrpmdb()
    try:
        indices = db.findbyname(pkg[0])
    except:
        indices = []

    h = None
    for i in indices:
        h = db[i]

    if h == None:
        return None
    
    try:
        pkgName = "%s-%s-%s" % (h['name'],
                                h['version'],
                                h['release'])
        log.log_me("getAdvisoryInfo for %s" % pkgName)
        info = s.errata.GetByPackage(pkgName,
                                     getVersion())
    except KeyboardInterrupt:
        raise CommunicationError(_("Connection aborted by the user"))
    except xmlrpclib.ResponseError:
        raise CommunicationError("Broken response from the server.")
    except xmlrpclib.Fault, f:
        print f
        return None
    
    if info:
        return info
    
    try:
        log.log_me("getAdvisoryInfo for %s-0-0" % h['name'])
        info = s.errata.GetByPackage("%s-0-0" % h['name'],
                                     getVersion())
    except KeyboardInterrupt:
        raise CommunicationError(_("Connection aborted by the user"))
    except xmlrpclib.ResponseError:
        raise CommunicationError("Broken response from the server.");
    except xmlrpclib.Fault, f:
        print f
        return None
    
    return info
    

def remoteAddPackages(pkg):
    log.log_me("Adding packages to package profile: %s" % pprint_pkglist(pkg))
    s = getServer()
    try:
        if type(pkg[0]) == type([]):
            #s.set_transport(2, 2)
            s.registration.add_packages(getSystemId(), pkg)
        else:
            s.registration.add_packages(getSystemId(), [pkg])            
    except KeyboardInterrupt:
        raise CommunicationError(_("Connection aborted by the user"))
    except (socket.error, socket.sslerror), e:
        if len(e.args) > 1:
            raise CommunicationError(e.args[1])
        else:
            raise CommunicationError(e.args[0])
    except cgiwrap.ProtocolError, e:
        raise CommunicationError(e.errmsg)
    except xmlrpclib.ResponseError:
        raise CommunicationError("Broken response from the server.");
    except xmlrpclib.Fault, f:
        raise CommunicationError(f.faultString)


def remoteDelPackages(pkg):
    log.log_me("Removing packages from package profile: %s" % pprint_pkglist(pkg))
    s = getServer()

    try:
        if type(pkg[0]) == type([]):
            #s.set_transport(2, 2)
            s.registration.delete_packages(getSystemId(), pkg)
        else:
            s.registration.delete_packages(getSystemId(), [pkg])            
    except KeyboardInterrupt:
        raise CommunicationError(_("Connection aborted by the user"))
    except (socket.error, socket.sslerror), e:
        if len(e.args) > 1:
            raise CommunicationError(e.args[1])
        else:
            raise CommunicationError(e.args[0])
    except cgiwrap.ProtocolError, e:
        raise CommunicationError(e.errmsg)
    except xmlrpclib.ResponseError:
        raise CommunicationError("Broken response from the server.");
    except xmlrpclib.Fault, f:
        raise CommunicationError(f.faultString)


def updatePackageProfile():
    log.log_me("Updating package profile")
    s = getServer()
    
    try:
        #s.set_transport(2, 2)
        s.registration.update_packages(getSystemId(),
                                       getInstalledPackageList())
    except KeyboardInterrupt:
        raise CommunicationError(_("Connection aborted by the user"))
    except (socket.error, socket.sslerror), e:
        if len(e.args) > 1:
            raise CommunicationError(e.args[1])
        else:
            raise CommunicationError(e.args[0])
    except cgiwrap.ProtocolError, e:
        raise CommunicationError(e.errmsg)
    except xmlrpclib.ResponseError:
        raise CommunicationError("Broken response from the server.");
    except xmlrpclib.Fault, f:
        if f.faultCode == 99:
            raise DelayError(f.faultString)
        else:
            raise CommunicationError(f.faultString)


def getInstalledPackageList(msgCallback = None, progressCallback = None):
    list = []

    if msgCallback != None:
        msgCallback(_("Getting list of packages installed on the system"))

    if cfg.readEntry("dbpath"):
        dbpath = cfg.readEntry("dbpath")
    else:
        dbpath = "/"

    db = openrpmdb()

    if progressCallback != None:
        count = 1
        total = 0

        if "firstkey" in dir(db):
            key = db.firstkey()
            h = db[key]
        else:
            iterator = db.match(0)
            h = iterator.next()
        while (h):
            count = count + 1
            if "nextkey" in dir(db):
                key = db.nextkey(key)
                if key:
                    h = db[key]
                else:
                    h = None
            else:
                h = iterator.next()

        total = count
        count = 1

    if "firstkey" in dir(db):
        key = db.firstkey()
        h = db[key]
    else:
        iterator = db.match(0)
        h = iterator.next()
    while (h):
        name = h['name']
        epoch = h['epoch']
        if epoch == None:
            epoch = ""
        version = h['version']
        release = h['release']

        list.append([name, version, release, epoch])
        if progressCallback != None:
            progressCallback(count, total)
            count = count + 1
        if "nextkey" in dir(db):
            key = db.nextkey(key)
            if key:
                h = db[key]
            else:
                h = None
        else:
            h = iterator.next()

    list.sort()
    return list

# returns new pkgList
def removeInstalledPackagesFromList(pkgList, pkgNames,
                                    msgCallback = None,
                                    progressCallback = None):
    
    removedList = []
    pkgCount = len(pkgList)
    installedPkgList = getInstalledPackageList()

    if msgCallback != None:
        msgCallback(_("Removing installed packages from list of updates"))


    # pretend the pkgNames are installed, with versions 0-0:0
    # is this sane? it sure makes life easier...
    for pkg in pkgNames:
        if not pkg in installedPkgList[0]:
            installedPkgList.append([pkg, '0.0','0.0',''])


    updatedList = []
    for index in range(pkgCount):
        # find all the installed packages in the pkglist
        for installedPkg in installedPkgList:
            if installedPkg[0] == pkgList[index][0]:
                ret = comparePackages(pkgList[index], installedPkg)
                if ret == 0:
                    # already installed
                    # make sure it wasn't added to the list already,
                    # which can happen when multiple packages with the
                    # same name are installed.
                    try:
                        del updatedList[map(lambda p: p[0], updatedList).index(installedPkg[0])]
                    except:
                        pass
                elif ret < 0:
                    # newer version already installed
                    # make sure it wasn't added to the list already,
                    # which can happen when multiple packages with the
                    # same name are installed.
                    try:
                        del updatedList[map(lambda p: p[0], updatedList).index(installedPkg[0])]
                    except:
                        pass
                else:
                    # we need to fetch it. avoid including multiple copies;
                    # sometimes two packages w/the same name are installed
                    # on a system.
                    if pkgList[index] not in updatedList:
                        updatedList.append(pkgList[index])

        if progressCallback:
            progressCallback(index+1, pkgCount)

    return updatedList

    
# side effects on pkgList
def removeSkipPackagesFromList(pkgList, msgCallback = None,
                               progressCallback = None):
    removedList = []
    pkgCount = len(pkgList)

    if cfg.readEntry('forceInstall'):
        return removedList
    
    if msgCallback != None:
        msgCallback(_("Removing packages marked to skip from list"))

    pkgSkipList = cfg.readEntry("pkgSkipList")
    for index in range(pkgCount):
        for pattern in pkgSkipList:
            if fnmatch.fnmatch(pkgList[index][0], pattern):
                removedList.append((pkgList[index], _("Pkg name/pattern")))
                break

        if progressCallback != None:
            progressCallback(index+1, pkgCount)

    for pkg in removedList:
        pkgList.remove(pkg[0])

    # just do a side effect on pkgList here, and return removedList
    # since we actually need the results of it
    return removedList


# check to see if a new file has a different MD5 sum
# from the package already on disk, and if true, if
# the on-disk file has been modified.
# this is a lot of args, but it helps...
def checkModified(index, fileNames, fileMD5s,installedFileNames,installedFileMD5s):
    ret = 0
    fileName = fileNames[index]

    MD5Res = ""

    #this is a little ugly, but the order of the filelist in the old and the new
    # pacakges arent the same.
    for j in range(len(installedFileNames)):
        if (installedFileNames[j] == fileName):
            if installedFileMD5s[j] == fileMD5s[index]:
                # the md5 of the file in the local db  is the same as the one in the
                # incoming package, so skip the rest of the check and return 0
                continue
	    # grrr, symlinks marked as config files pointing to dirs are baaad, okay?
	    if not os.path.isdir(fileName):
            	if installedFileMD5s[j] != md5(fileName):
                    # the local file is different than the file in the local db
                    ret = 1
            	else:
                    # not changed on disk
                    pass
            break
    return ret


# check if the signature matches the redhat.com one
def hasBadSignature(pkg):
    hdr = getHeader(pkg)
    fileName = "%s/%s.%s.rpm" % (cfg.readEntry("storageDir"),
                                 pkgToString(pkg),
                                 hdr['arch'])
    # must use the actual header from the real package.  The headers
    # the server returns by themselves have no signature info in them.
    fd = os.open(fileName, os.O_RDONLY)
    try:
        hdr = rpm.headerFromPackage(fd)[0]
    except:
        raise RpmError(_("Failed to read header from %s for signature verification") % fileName)
    
    os.close(fd)

    try:
        # only newer RPMs allow this
        if hdr['SIGGPG'] == None:
            return 1
        if "checksig" in dir(rpm):
            # RPM spews to stdout/stderr.  Redirect.
            saveStdout = os.dup(1)
            saveStderr = os.dup(2)
            redirStdout = os.open("/dev/null", os.O_WRONLY | os.O_APPEND)
            redirStderr = os.open("/dev/null", os.O_WRONLY | os.O_APPEND)
            os.dup2(redirStdout, 1)
            os.dup2(redirStderr, 2)
            # now do the rpm thing
            ret = rpm.checksig(fileName, rpm.RPMTAG_SIGGPG)
            # restore normal stdout and stderr
            os.dup2(saveStdout, 1)
            os.dup2(saveStderr, 2)
            # clean up
            os.close(redirStdout)
            os.close(redirStderr)
            os.close(saveStdout)
            os.close(saveStderr)
            if ret:
                return 2

        return 0
    except KeyError:
        # legacy 3.0.x RPM
        sigOk = 1
        lines = os.popen("/bin/rpm -Kv %s 2>&1" % fileName, "r").read()
        if re.search("BAD signature from \"Red Hat, Inc <security@redhat.com>\"", lines):
            sigOk = 2
        elif re.search("Good signature from \"Red Hat, Inc <security@redhat.com>\"", lines) and re.search("Fingerprint: CA20 8686 2BD6 9DFC 65F6  ECC4 2191 80CD DB42 A60E", lines):
            sigOk = 0

        return sigOk

# trim down pkgList to just the files in pkgNames, ie the ones
# specified on the commandline. ugly function name, but it sorta
# follows the existing scheme. We really should just be requesting
# info on the file specified on the commandline, but this works for now.
#
# returns the new pkgList
def removeNonPkgNamesPackagesFromList(pkgList,pkgNames, msgCallback = None,
                                      progressCallback = None):
    removedList = []
    pkgCount = len(pkgList)
    pkgNamesCount = len(pkgNames)
    
    if msgCallback != None:
        msgCallback(_("Removing packages with files not specified from list"))

    tmp_pkgList = []
    for pkg in pkgList:
        if (pkg[0] in pkgNames):
            tmp_pkgList.append(pkg)

    return tmp_pkgList


def checkHeaderForFileConfigExcludes(h,package,db):
    fflag = 0

    # might as well just do this once...
    fileNames = h['filenames']
    fileMD5s = h['filemd5s']
    fileFlags = h['fileflags']

    installedHdr = installedHeader(h['name'],db)
    if not installedHdr:
        #throw a fault? should this happen?
        return None
        
    installedFileNames = installedHdr['filenames']
    installedFileMD5s = installedHdr['filemd5s']
    
    removedList = []
    if cfg.readEntry('forceInstall'):
        return None
    
    # bleah, have to use the index here because
    # of rpm's storing of filenames and their md5sums in parallel lists
    for f_i in range(len(fileNames)):
        # code to see if we want to disable this
        for pattern in cfg.readEntry("fileSkipList"):
            if fnmatch.fnmatch(fileNames[f_i],pattern):
                # got to get a better string to use here
                removedList.append((package, _("File Name/pattern")))
                fflag = 1
                break
            # if we found a matching file, no need to
            # examine the rest in this package         
            if fflag:
                break

    if cfg.readEntry("configFilesToIgnore"):
        configFilesToIgnore = cfg.readEntry("configFilesToIgnore")
    else:
        configFilesToIgnore = []

    for f_i in range(len(fileNames)):
        # Deal with config files
        if cfg.readEntry("noReplaceConfig"):
            # check for files that are config files, but skips those that
            # arent going to be replaced anyway
            # (1 << 4) == rpm.RPMFILE_NOREPLACE if it existed
            if fileFlags[f_i] & rpm.RPMFILE_CONFIG and not \
               fileFlags[f_i] & (1 << 4):
                if fileNames[f_i] not in configFilesToIgnore:
                    # check if config file and if so, if modified
                    if checkModified(f_i, fileNames, fileMD5s,
                                     installedFileNames, installedFileMD5s):
                        removedList.append((package, _("Config modified")))
                        fflag = 1
                        break

        if fflag:
            break     

    if len(removedList):
        return removedList
    else:
        return None

# side effects on pkgList,
def removeSkipFilesPackagesFromList(pkgList, msgCallback = None,
                                    progressCallback = None):
    removedList = []

    pkgCount = len(pkgList)

    if msgCallback != None:
        msgCallback(_("Removing packages with files marked to skip from list"))

    p_i = 0
    h_list = getHeaders(pkgList, progressCallback=progressCallback)

    # need the db for the checkHeaderFiles
    db = openrpmdb()
    
    if not cfg.readEntry('forceInstall'):
	    for i in h_list:
   	     tmp_list = 0
             #check the files based on the new header
   	     tmp_list = checkHeaderForFileConfigExcludes(i,pkgList[p_i],db)
	     p_i = p_i + 1
   	     if tmp_list:
   	         removedList = removedList + tmp_list

    # yeah, this is ugly, but look ma! lamdba/map magic!
    # make sure the lists are uniq, and make sure we remove the right packages
    filteredRemovedList = map(lambda x: x[0], removedList)
    blip = map(lambda x: ("%s-%s-%s-%s" % (x[0][0],x[0][1],x[0][2],x[0][3]), x) , removedList)

    d = {}
    for i in blip: d[i[0]] = i
    blargh = d.values()

    filteredRemovedList = map(lambda x: x[1][0], blargh)
    removedList = map(lambda x: x[1], blargh)

    for pkg in filteredRemovedList:
	if pkg in pkgList:
            pkgList.remove(pkg)

    return removedList

# return a list of packages that need to be updated according
# to the preferences set in the config file, etc.
def getUpdatedPackageList(pkgNames, msgCallback = None,
                          progressCallback = None, refreshCallback = None,availList=None):
    pkgSkipList = []
    
    if msgCallback != None:
        msgCallback(_("Retrieving list of all available packages"))

    if not len(pkgNames):
        # dont fetch a listall...

        if availList:
            pkgList = availList
        else:
            pkgList = getAvailablePackageList()
        
        
    	pkgList = removeInstalledPackagesFromList(pkgList, [],
                                                  msgCallback,
                                                  progressCallback)
        
    else:
        # need a new xmlrpc call to just get info on the packages specified
        # it could reduce bandwidth a bit in exchange for more dynamic
        # content serverside
        if availList:
            pkgList = availList
        else:
            pkgList = getAvailablePackageList()
        
	pkgList = removeInstalledPackagesFromList(pkgList, pkgNames,
                                                  msgCallback,
                                                  progressCallback)
        
        pkgList = removeNonPkgNamesPackagesFromList(pkgList,pkgNames,
                                                    msgCallback,
                                                    progressCallback)
        
    if len(pkgList) > 0:
        pkgSkipList = removeSkipPackagesFromList(pkgList, msgCallback,
                                                 progressCallback)
        
    if len(pkgList) > 0:
        if msgCallback != None:
            msgCallback(_("Getting headers for available packages"))
        total = len(pkgList)
        
        h_list = getHeaders(pkgList, msgCallback, progressCallback,
                            refreshCallback)

    if len(pkgList) > 0:
        pkgSkipList = pkgSkipList + removeSkipFilesPackagesFromList(pkgList, msgCallback,
                                                                    progressCallback)
            
    if len(pkgSkipList) > 0:
        if msgCallback != None:
            msgCallback(_("Getting headers for skipped packages"))
        total = len(pkgSkipList)

        for index in range(total):
            hdr = getHeader(pkgSkipList[index][0])
            headerCache[pkgToString(pkgSkipList[index][0])] = hdr
            if progressCallback:
                progressCallback(index+1, total)

    pkgList.sort()
    pkgSkipList.sort()
    return (pkgList, pkgSkipList)

    
def totalSize(pkgList):
    total = 0
    
    for pkg in pkgList:
        if len(pkg) < 5:
            continue
        total = total + pkg[4]

    return total
    
def freeDiskSpace():
    import statvfs

    dfInfo = os.statvfs(cfg.readEntry("storageDir"))
    return long(dfInfo[statvfs.F_BAVAIL]) * (dfInfo[statvfs.F_BSIZE])


def rpmCallback(what, amount, total, key, cb):
    if what == rpm.RPMCALLBACK_INST_OPEN_FILE:
        pass
        #fd = os.open(key, os.O_RDONLY)
        #return fd
    elif what == rpm.RPMCALLBACK_INST_START:
        pass
    elif what == rpm.RPMCALLBACK_INST_CLOSE_FILE:
        print
    elif what == rpm.RPMCALLBACK_TRANS_PROGRESS:
        if cb:
            cb(amount, total)
        else:
            print "transaction %.5s%% done\r" % ((float(amount) / total) * 100),
    elif what == rpm.RPMCALLBACK_INST_PROGRESS:
        print "installation %.5s%% done\r" % ((float(amount) / total) * 100), 


def showSkipListHits(skiplist):
    errmsg = _("""
    To solve all dependencies for the RPMs you have selected, The following
    packages you have marked to exclude would have to be added to the set:
    """)
    errmsg = errmsg + _("\n    %-30.30s\t%-30.30s\n") % (
        _("Package Name"), _("Reason For Skipping"))
    errmsg = errmsg + " " * 4 +  "=" * 70 + "\n"
    for i in skiplist:
        package = "%s-%s-%s" % (i[0][0], i[0][1], i[0][2])
        errmsg = errmsg + "    %-30.30s\t%-30.30s\n" % (package, i[1])
    return errmsg

# This class is the engine of the dependency solver so that the external
# interface to dryRun remains readable
class DependencySolver:
    def __init__(self, selected):
        # turn on debugging for this particular class here
        self.debug = 0
        self.selected = selected
        available = getAvailablePackageList()

        # exclude the already selected ones from the available list
        xlist = map(lambda a: a[0], self.selected)
        self.available = []

        # a list of excludes that might break our dependency chains
        self.excludes = []

        # make sure the ones that are skipped are not in here
        available, junk = self.__skip(available)
        for i in range(len(available)):
            if not available[i][0] in xlist:
                self.available.append(available[i])

        # placeholder for the database and transaction set
        self.db = self.ts = None
        
        # null callbacks for now
        self.msgCallback = self.progressCallback = self.refreshCallback = None
        
    # set the callbacks
    def set_callback(self, name, function):
        if name == "msg":
            self.msgCallback = function
        elif name == "progress":
            self.progressCallback = function
        elif name == "refresh":
            self.refreshCallback = function

    # get ready to rumble
    def setup(self):
        # prepare for the transaction list
        self.db = openrpmdb()
        self.ts = rpm.TransactionSet("/", self.db)

        if self.msgCallback != None:
            self.msgCallback(_("Testing package set / solving RPM inter-dependencies"))
        pkgset = self.selected
        self.selected = [] # the ___add will fill it back as it does ts.add
        self.__add(pkgset, check = 0)
        self.ts.order()

    # refresh the screen or something
    def __refresh(self):
        if self.refreshCallback:
            self.refreshCallback()
        
    # sanitize a list of packages for the ones we have to skip
    def __skip(self, plist):
        # make sure the new packages don't include any packages
        # with packages excludes
        plist.sort()
        xlist = removeSkipPackagesFromList(plist)
        for x in xlist:
            # remove the package
            if self.debug:
                print "Removing package", x
            if x[0] in self.available: # make it unavailable too
                self.available.remove(x[0])
            self.excludes.append(x)
        return plist, xlist

    # add packages to a transaction set
    def __add(self, plist, check = 0):
        if self.debug:
            print "Candidates for the selected list:", plist
        # get the headers
        h_list = getHeaders(plist, refreshCallback = self.refreshCallback)

        added = [] # what packages do we add?
        
        # add them one by one
        counter = -1
        for h in h_list:
            counter = counter + 1
            if not h:
                continue
            if check:
                xlist = checkHeaderForFileConfigExcludes(h, plist[counter], self.db)
                if xlist: # those packages have to be skipped too
                    if self.debug:
                        print "Got exclude list from config file check:", xlist
                    for x in xlist:
                        if self.debug:
                            print "Skipping package %s because of config file change" % (x,)
                        if x[0] in self.available:
                            self.available.remove(x[0])
                        self.excludes.append(x)
                    continue
            newp = plist[counter]
            if newp not in self.selected:
                self.selected.append(newp)
                if self.debug:
                    print "Adding %s to transaction set" % newp
                self.ts.add(h, newp, "u")
                added.append(newp)
            if newp in self.available: # no longer available
                self.available.remove(plist[counter])
            self.__refresh()
        return added
    
    # Solve dependencies
    def __dependencies(self, dependencies):
        if not dependencies:
            return 0
        changed = 0
        if self.debug:
            print "Dependencies:", dependencies
        solved = solveDep(map(lambda a: a[1], dependencies),
                          refreshCallback = self.refreshCallback)
        if self.debug:
            print "Got back response:", solved
        # If we got a response back, take out the packages that have been
        # already selected so we don't loop without a purpose
        for s in solved[:]:
            if s in self.selected:
                if self.debug:
                    print "Already selected:", s
                solved.remove(s)
        # if we did not get a solution pack back, try looking at the
        # problem packages that are thorwing these dependencies.
        if not solved:
            solved = []
            snames = map(lambda a: a[0], dependencies)
            for a in self.available:
                if a[0] in snames:
                    if self.debug:
                        print "Package %s raised invalid dependency. Adding %s to set" % (
                        a[0], a)
                    solved.append(a)
                    changed = 1
        # sanitize this list
        solved, skiplist = self.__skip(solved)
        added = self.__add(solved, check = 1)
        if added:
            changed = 1
        return changed
        
    # Solve conflicts
    def __conflicts(self, conflicts):        
        if not conflicts:
            return 0
        changed = 0
        if self.debug:
            print "Conflicts:", conflicts
        solved = []
        for c in conflicts:
            # if one of these packages is still available, grab it
            for a in self.available:
                if a[0] == c[0] or a[0] == c[1]:
                    solved.append(a)
                    changed = 1
        if solved is None and not changed:
            raise DependencyError(_("Could not satisfy conflict dependencies."))
        solved, skiplist = self.__skip(solved)
        added = self.__add(solved, check = 1)
        if added:
            changed = 1
        return changed
    
    # Format the dependency error nicely for the plain guy
    def __dep_error(self, deps):
        s = StringIO.StringIO()
        for dep in deps:
            ((name, version, release), (needsName, needsVersion),
             flags, suggested, sense) = dep
            if needsVersion:
                s.write(_("%-40.40s requires %s\n") % (
                    "%s-%s-%s" % (name, version, release),
                    "%s-%s" % (needsName, needsVersion)))
            else:
                s.write(_("%-40.40s requires %s\n") % (
                    "%s-%s-%s" % (name, version, release),
                    needsName))
        ret = s.getvalue()
        s.close()
        return ret
        
    # generic dependency handler
    def process_deps(self, deps):
        if not deps:
            return 0
        conflicts = []
        dependencies = []
        last_excludes = self.excludes
        self.excludes = []
        for dep in deps:
            if self.debug:
                print "Processing dependency", dep
            ((name, version, release), (needsName, needsVersion),
             flags, suggested, sense) = dep
            if sense == rpm.RPMDEP_SENSE_REQUIRES:
                dependencies.append(name, needsName)
            elif sense == rpm.RPMDEP_SENSE_CONFLICTS:
                conflicts.append(name, needsName)
            else:
                raise DependencyError(_("Don't know how to handle dependency sense \"%s\"") %
                                      sense)
            self.__refresh()
        # now solve the dependencies...
        changed = self.__dependencies(dependencies)
        self.__refresh()
        # and the conflicts...
        changed = changed + self.__conflicts(conflicts)
        self.__refresh()
        if not changed: # blow up
            if not self.excludes: # get the last round instead
                self.excludes = last_excludes
            errmsg = ""
            if self.excludes:
                errmsg = showSkipListHits(self.excludes)
            dep_error = self.__dep_error(deps)
            raise DependencyError(_("%s\nUnresolvable chain of dependencies:\n"
                                        "%s") % (errmsg, dep_error))
        return 1
    
    # run a dependency iteration
    def solvedep(self):
        if self.debug:
            print "\nChecking for dependencies"        
        self.__refresh()
        deps = self.ts.depcheck()       
        if self.debug and deps:
            print "RPM returned %d deps." % len(deps)
        if deps:
            # make the deps list unique:
            deps.sort()
            last = deps[-1]
            for i in range(len(deps)-2, -1, -1):
                if last == deps[i]:
                    del deps[i]
                else:
                    last = deps[i]
        ret = self.process_deps(deps)
        return ret

    # run the transaction set now to check for any other problems
    def run(self):
        if self.debug:
            print "Running transaction (final step)..."
        self.ts.order()
        self.__refresh()
        ret = self.ts.run(rpm.RPMTRANS_FLAG_BUILD_PROBS, 0,
                          rpmCallback, self.progressCallback)
        self.__refresh()
        return ret

# The new dryRun
def dryRun(pkgList, msgCallback = None, progressCallback = None, refreshCallback=None):
    # XXX: initialize correctly from the --verbose cmd line flag.
    debug = 0
    
    if debug:
        print "Called dryRun:", pkgList

    depsolve = DependencySolver(pkgList)

    # set the callbacks wwe need
    depsolve.set_callback("msg", msgCallback)
    depsolve.set_callback("progress", progressCallback)
    depsolve.set_callback("refresh", refreshCallback)
    
    # set up the trabsaction set & all
    depsolve.setup()
    while 1:
        if debug:
            print "\nSelected so far", depsolve.selected
        ret = depsolve.solvedep()
        if debug:
            print "Got code %d from solvedep" % ret
        if ret == 0:
            break
        elif ret < 0: # not really reached
            raise DependencyError(_("Could not complete dependency resolution"))
    rc = depsolve.run()
    if progressCallback:
        progressCallback(1, 1)       
    if len(rc):
        msg = ""
        for code in rc:
            msg = msg + code[0] + "\n"
            raise ConflictError(_("Test install failed because of package conflicts:\n") + msg)
    newList = []
    for s in depsolve.selected:
        if not s in pkgList:
            pkgList.append(s)
            newList.append(s)
    return (newList, [])

def pprint_pkglist(pkglist):
    if type(pkglist) == type([]):
        foo = map(lambda a : "%s-%s-%s" % (a[0],a[1],a[2]), pkglist)
    else:
        foo = "%s-%s-%s" % (pkglist[0], pkglist[1], pkglist[2])
    return foo
                 
def installPackages(pkgList, rpmCallback):
    try:
        log.log_me("installing packages: %s " % pprint_pkglist(pkgList))
    except:
        pass

    # a list of all the packge names to treat as kernels.
    # "hardcode" it here so people dont mess with it as much
    total = len(pkgList)

    oldInstalledPackages = getInstalledPackageList()
    
    # delete any existing db, open in write mode
    db = openrpmdb(option=1)
    ts = rpm.TransactionSet("/", db)

    kernels = []
    for p_i in range(total):
        hdr = getHeader(pkgList[p_i])
        if hdr:
            if "kernel" in hdr['provides']:
                ts.add(hdr,hdr, "i")
                kernels.append(hdr)
            else:
                ts.add(hdr, hdr, "u")
        else:
            raise RpmError(_("Could not find header for %s") % pkgList[p_i])

    ts.order()
    deps = ts.depcheck()
    if deps:
        raise DependencyError(_("Dependencies should have already been resolved, but they are not."))

    if cfg.readEntry("justdb"):
        rc = ts.run(rpm.RPMTRANS_FLAG_JUSTDB, 0,
                    rpmCallback, cfg.readEntry("storageDir"))
    else:
        rc = ts.run(0, 0, rpmCallback, cfg.readEntry("storageDir"))
    if rc:
        errors = "\n"
        for e in rc:
            try:
                errors = errors + e[1] + "\n"
            except:
                error = errors + e + "\n"
        raise RpmError(_("Failed installing packages: %s") % errors)

    newInstalledPackages = getInstalledPackageList()

    removedPackages = []
    addedPackages = []

    # find packages that were removed
    for pkg in oldInstalledPackages:
        if pkg not in newInstalledPackages:
            removedPackages.append(pkg)

    for pkg in newInstalledPackages:
        if pkg not in oldInstalledPackages:
            addedPackages.append(pkg)

    if len(removedPackages):
    	remoteDelPackages(removedPackages)
    if len(addedPackages):    
    	remoteAddPackages(addedPackages)

    return kernels

def installBootLoader(kernels):
    
    # option to disable this permanently for those using grub,etc
    if cfg.readEntry("noBootLoader"):
        return

    # dont run the new lilo if were running "justdb"
    if cfg.readEntry("justdb"):
        return
    
    if archGlob() == 'i?86':
        try:
            import lilocfg
        except ImportError:
            # if lilocfg isnt found, bale...
            pass

        log.log_me("Modifying lilo.conf to include the new kernel info")
        version_list = []
        default = 0
        kernelList = []
        for krn in kernels:
            if "kernel-BOOT" in krn['provides']:
                extraInfo = "BOOT"
            elif "kernel-smp" in krn['provides']:
                extraInfo = "smp"
            elif "kernel-enterprise" in krn['provides']:
                extraInfo = "enterprise"
            else:
                extraInfo = None

            if extraInfo == None:
                infoString = ""
            else:
                infoString = extraInfo
            verRel = "%s-%s%s" % (krn['version'], krn['release'],infoString)
            kernelList.append((verRel,extraInfo))
            log.log_me("Adding %s to lilo.conf" % verRel)

        log.log_me("Running lilo with the new configuration")
        ret = lilocfg.installNewImages(kernelList,test=0)
                        
def msgCallback(msg):
    print msg

def simpleCallback(count, total):
    print "%.5s%% done\r" % ((float(count) / total) * 100),
